@@ -4,13 +4,12 @@ from functools import lru_cache
 import logging
 import pathlib
 
-from adcircpy import AdcircMesh, AdcircRun, Tides, server
+from adcircpy import AdcircMesh, AdcircRun, server, Tides
 
 logger = logging.getLogger(__name__)
 
 
 class AdcircCommand:
-
     def __init__(self, args):
         self.args = args
 
@@ -20,16 +19,17 @@ class AdcircCommand:
         if self.args.generate_only:
             logger.info('Generate only is active, writting to disk.')
             self.driver.write(
-                    self.args.output_directory,
-                    overwrite=self.args.overwrite
+                self.args.output_directory,
+                nproc=self.args.nproc,
+                overwrite=self.args.overwrite,
             )
             return
 
         outputs = self.driver.run(
-                outdir=self.output_directory,
-                nproc=self.args.nproc,
-                overwrite=self.args.overwrite,
-                server_config=self.server_config,
+            outdir=self.output_directory,
+            nproc=self.args.nproc,
+            overwrite=self.args.overwrite,
+            server_config=self.server_config,
         )
         self._output_collection = outputs
 
@@ -37,11 +37,11 @@ class AdcircCommand:
     @lru_cache(maxsize=None)
     def driver(self):
         driver = AdcircRun(
-                self.mesh,
-                self.start_date,
-                self.end_date,
-                self.spinup_time,
-                server_config=self.server_config
+            self.mesh,
+            self.start_date,
+            self.end_date,
+            self.spinup_time,
+            server_config=self.server_config,
         )
         self._enable_outputs(driver)
         if self.args.timestep:
@@ -121,13 +121,11 @@ class AdcircCommand:
                 _all = (*_major, 'Mm', 'Mf', 'M4', 'MN4', 'MS4', '2N2', 'S1')
             elif self.args.tidal_database == 'hamtide':
                 _all = _major
-            if ('all' in self.args.constituents
-                    and len(self.args.constituents) > 1):
+            if 'all' in self.args.constituents and len(self.args.constituents) > 1:
                 msg = 'When using all, must only pass one'
                 raise IOError(msg)
 
-            elif ('major' in self.args.constituents
-                  and len(self.args.constituents) > 1):
+            elif 'major' in self.args.constituents and len(self.args.constituents) > 1:
                 msg = 'When using major, must only pass one'
                 raise IOError(msg)
             if 'all' in self.args.constituents:
@@ -143,39 +141,37 @@ class AdcircCommand:
     @lru_cache(maxsize=None)
     def server_config(self):
         if self.args.hostname:
-            if (not self.args.use_slurm or
-                    not self.args.use_torque or
-                    not self.args.use_pbs):
+            if not self.args.use_slurm or not self.args.use_torque or not self.args.use_pbs:
                 return server.ServerConfig(
-                        hostname=self.args.hostname,
-                        nprocs=self.args.nproc,
-                        wdir=self.args.wdir,
-                        binaries_prefix=self.args.binaries_prefix,
-                        source_script=self.args.source_script,
-                        additional_mpi_options=self.args.additional_mpi_options,
+                    hostname=self.args.hostname,
+                    nprocs=self.args.nproc,
+                    wdir=self.args.wdir,
+                    binaries_prefix=self.args.binaries_prefix,
+                    source_script=self.args.source_script,
+                    additional_mpi_options=self.args.additional_mpi_options,
                 )
 
         if self.args.use_slurm:
             kwargs = {
-                "account": self.args.account,
-                "ntasks": self.args.slurm_ntasks,
-                "partition": self.args.partition,
-                "walltime": timedelta(hours=self.args.walltime),
-                "mail_type": self.args.mail_type,
-                "mail_user": self.args.mail_user,
-                "log_filename": self.args.log_filename,
-                "modules": self.args.modules,
-                "path_prefix": self.args.binaries_prefix,
-                "extra_commands": self.args.extra_commands,
-                "launcher": self.args.slurm_launcher,
-                "nodes": self.args.slurm_nodes
+                'account': self.args.account,
+                'ntasks': self.args.slurm_ntasks,
+                'partition': self.args.partition,
+                'walltime': timedelta(hours=self.args.walltime),
+                'mail_type': self.args.mail_type,
+                'mail_user': self.args.mail_user,
+                'log_filename': self.args.log_filename,
+                'modules': self.args.modules,
+                'path_prefix': self.args.binaries_prefix,
+                'extra_commands': self.args.extra_commands,
+                'launcher': self.args.slurm_launcher,
+                'nodes': self.args.slurm_nodes,
             }
             if self.args.slurm_filename is not None:
-                kwargs.update({"filename": self.args.slurm_ntasks})
+                kwargs.update({'filename': self.args.slurm_ntasks})
             if self.args.slurm_rundir is not None:
-                kwargs.update({"run_directory": self.args.slurm_rundir})
+                kwargs.update({'run_directory': self.args.slurm_rundir})
             if self.args.run_name is not None:
-                kwargs.update({"run_name": self.args.run_name})
+                kwargs.update({'run_name': self.args.run_name})
 
             return server.SlurmConfig(**kwargs)
 
@@ -190,25 +186,21 @@ class AdcircCommand:
         self._init_output_stations(driver)
 
     def _enable_output(self, driver, name, _type):
-        fs = getattr(self.args, f"{name}_{_type}_sampling_rate")
+        fs = getattr(self.args, f'{name}_{_type}_sampling_rate')
         if fs is not None:
             fs = timedelta(minutes=fs)
-        fss = getattr(self.args, f"{name}_{_type}_sampling_rate_spinup")
+        fss = getattr(self.args, f'{name}_{_type}_sampling_rate_spinup')
         if fss is not None:
             fss = timedelta(minutes=fss)
-        ha = getattr(self.args, f"{name}_{_type}_harmonic_analysis")
+        ha = getattr(self.args, f'{name}_{_type}_harmonic_analysis')
         # has = getattr(self.args, f"{name}_{_type}_harmonic_analysis_spinup")
-        getattr(driver, f"set_{name}_{_type}_output")(
-                sampling_rate=fs,
-                harmonic_analysis=ha,
-                spinup=fss,
-                netcdf=self.args.netcdf,
+        getattr(driver, f'set_{name}_{_type}_output')(
+            sampling_rate=fs, harmonic_analysis=ha, spinup=fss, netcdf=self.args.netcdf,
         )
 
     def _init_output_stations(self, driver):
         if self.args.stations_file is not None:
-            driver.import_stations(
-                    pathlib.Path(self.args.stations_file).resolve())
+            driver.import_stations(pathlib.Path(self.args.stations_file).resolve())
             self._enable_output(driver, 'elevation', 'stations')
             self._enable_output(driver, 'velocity', 'stations')
             self._enable_output(driver, 'meteorological', 'stations')
@@ -217,23 +209,18 @@ class AdcircCommand:
     @property
     @lru_cache(maxsize=None)
     def _mesh(self):
-        mesh = AdcircMesh.open(
-                self.args.mesh,
-                self.args.crs
-        )
+        mesh = AdcircMesh.open(self.args.mesh, self.args.crs)
 
         if self.args.generate_boundaries:
             mesh.generate_boundaries(
-                    threshold=self.args.boundaries_threshold,
-                    land_ibtype=self.args.land_ibtype,
-                    interior_ibtype=self.args.island_ibtype,
+                threshold=self.args.boundaries_threshold,
+                land_ibtype=self.args.land_ibtype,
+                interior_ibtype=self.args.island_ibtype,
             )
 
         # set nodal attributes
         if self.args.fort13 is not None:
-            mesh.import_nodal_attributes(
-                    pathlib.Path(self.args.fort13).resolve()
-            )
+            mesh.import_nodal_attributes(pathlib.Path(self.args.fort13).resolve())
 
         if 'all' in self.args.coldstart_attributes:
             for attr in mesh.get_nodal_attribute_names():
@@ -255,8 +242,7 @@ class AdcircCommand:
         if self.args.generate_linear_mannings is True:
             mesh.generate_linear_mannings_n()
         elif self.args.generate_constant_mannings is not None:
-            mesh.generate_constant_mannings_n(
-                self.args.generate_constant_mannings)
+            mesh.generate_constant_mannings_n(self.args.generate_constant_mannings)
 
         if self.tidal_forcing is not None:
             mesh.add_forcing(self.tidal_forcing)
