@@ -5,12 +5,7 @@ from adcircpy.forcing.winds.base import WindForcing
 
 
 class AtmosphericMeshForcing(WindForcing):
-    def __init__(
-            self,
-            filename: PathLike,
-            nws: int = 5,
-            interval_seconds: int = 3600
-    ):
+    def __init__(self, filename: PathLike, nws: int = 5, interval_seconds: int = 3600):
         if not isinstance(filename, Path):
             filename = Path(filename)
         self.filename = filename
