@@ -9,8 +9,10 @@ import logging
 import os
 from os import PathLike
 import pathlib
+import socket
 import time
-from typing import Any, Union
+from typing import Any, TextIO, Union
+from urllib.error import URLError
 import zipfile
 
 import appdirs
@@ -29,6 +31,7 @@ from shapely.geometry import Point, Polygon
 import utm
 
 from adcircpy.forcing.winds.base import WindForcing
+from adcircpy.plotting import plot_polygon, plot_polygons
 
 logger = logging.getLogger(__name__)
 
@@ -75,7 +78,7 @@ def plot_coastline(axis: Axes = None, show: bool = False, save_filename: PathLik
 
 
 class FileDeck(Enum):
-    """ http://hurricanes.ral.ucar.edu/realtime/ """
+    """http://hurricanes.ral.ucar.edu/realtime/"""
 
     a = 'a'
     b = 'b'
@@ -99,8 +102,11 @@ class VortexForcing:
         file_deck: FileDeck = None,
         mode: Mode = None,
         record_type: str = None,
+        filename: PathLike = None,
     ):
         self.__dataframe = None
+        self.__filename = None
+
         self.__atcf = None
         self.__storm_id = None
         self.__start_date = start_date  # initially used to filter A-deck here
@@ -115,6 +121,7 @@ class VortexForcing:
         self.file_deck = file_deck
         self.mode = mode
         self.record_type = record_type
+        self.filename = filename
 
         if isinstance(storm, DataFrame):
             self.dataframe = storm
@@ -124,18 +131,35 @@ class VortexForcing:
             if os.path.exists(storm):
                 self.__atcf = io.open(storm, 'rb')
             else:
-                self.storm_id = storm
+                try:
+                    self.storm_id = storm
+                except ValueError:
+                    if pathlib.Path(storm).exists():
+                        self.filename = storm
+                    else:
+                        raise
 
         self.__previous_configuration = {
             'storm_id': self.storm_id,
             'file_deck': self.file_deck,
             'mode': self.mode,
+            'filename': self.filename,
         }
 
         # use start and end dates to mask dataframe here
         self.start_date = start_date
         self.end_date = end_date
 
+    @property
+    def filename(self) -> pathlib.Path:
+        return self.__filename
+
+    @filename.setter
+    def filename(self, filename: PathLike):
+        if filename is not None and not isinstance(filename, pathlib.Path):
+            filename = pathlib.Path(filename)
+        self.__filename = filename
+
     def __str__(self):
         record_number = self.__generate_record_numbers()
         lines = []
@@ -144,7 +168,9 @@ class VortexForcing:
 
         for column in dataframe.select_dtypes(include=['float']):
             if column not in ['latitude', 'longitude']:
-                dataframe.loc[:, column] = dataframe[column].round(0).astype('Int64', copy=False)
+                dataframe.loc[:, column] = (
+                    dataframe[column].round(0).astype('Int64', copy=False)
+                )
 
         for i, (_, row) in enumerate(dataframe.iterrows()):
             line = []
@@ -231,24 +257,32 @@ class VortexForcing:
     def write(self, path: PathLike, overwrite: bool = False):
         if not isinstance(path, pathlib.Path):
             path = pathlib.Path(path)
-        if path.exists() and overwrite is False:
-            raise Exception('File exist, set overwrite=True to allow overwrite.')
-        with open(path, 'w') as f:
-            f.write(str(self))
+        if overwrite or not path.exists():
+            with open(path, 'w') as f:
+                f.write(str(self))
+        else:
+            logger.warning(f'skipping existing file "{path}"')
 
     @property
     def storm_id(self) -> str:
         if self.__storm_id is None and not self.__invalid_storm_name:
             if self.__dataframe is not None:
-                storm_id = get_atcf_id(
-                    storm_name=self.__dataframe['name'].tolist()[-1],
-                    year=self.__dataframe['datetime'].tolist()[-1].year,
+                storm_id = (
+                    f'{self.__dataframe["basin"].iloc[-1]}'
+                    f'{self.__dataframe["storm_number"].iloc[-1]}'
+                    f'{self.__dataframe["datetime"].iloc[-1].year}'
                 )
                 try:
-                    get_atcf_file(storm_id, self.file_deck, self.mode)
                     self.storm_id = storm_id
-                except:
-                    self.__invalid_storm_name = True
+                except ValueError:
+                    try:
+                        storm_id = get_atcf_id(
+                            storm_name=self.__dataframe['name'].tolist()[-1],
+                            year=self.__dataframe['datetime'].tolist()[-1].year,
+                        )
+                        self.storm_id = storm_id
+                    except ValueError:
+                        self.__invalid_storm_name = True
         return self.__storm_id
 
     @storm_id.setter
@@ -320,7 +354,7 @@ class VortexForcing:
 
     @property
     def duration(self) -> float:
-        d = (self.end_date - self.start_date).days
+        d = (self.end_date - self.start_date) / timedelta(days=1)
         return d
 
     @property
@@ -425,6 +459,7 @@ class VortexForcing:
             'storm_id': self.storm_id,
             'file_deck': self.file_deck,
             'mode': self.mode,
+            'filename': self.filename,
         }
 
         if (
@@ -442,6 +477,7 @@ class VortexForcing:
             'storm_id': self.storm_id,
             'file_deck': self.file_deck,
             'mode': self.mode,
+            'filename': self.filename,
         }
 
         # only download new file if the configuration has changed since the last download
@@ -450,153 +486,162 @@ class VortexForcing:
             or len(self.__dataframe) == 0
             or configuration != self.__previous_configuration
         ):
-            # https://www.nrlmry.navy.mil/atcf_web/docs/database/new/abdeck.txt
+            if configuration['filename'] is not None:
+                dataframe = read_atcf(configuration['filename'])
+            else:
+                # https://www.nrlmry.navy.mil/atcf_web/docs/database/new/abdeck.txt
 
-            columns = [
-                'basin',
-                'storm_number',
-                'datetime',
-                'record_type',
-                'latitude',
-                'longitude',
-                'max_sustained_wind_speed',
-                'central_pressure',
-                'development_level',
-                'isotach',
-                'quadrant',
-                'radius_for_NEQ',
-                'radius_for_SEQ',
-                'radius_for_SWQ',
-                'radius_for_NWQ',
-                'background_pressure',
-                'radius_of_last_closed_isobar',
-                'radius_of_maximum_winds',
-                'name',
-                'direction',
-                'speed',
-            ]
+                columns = [
+                    'basin',
+                    'storm_number',
+                    'datetime',
+                    'record_type',
+                    'latitude',
+                    'longitude',
+                    'max_sustained_wind_speed',
+                    'central_pressure',
+                    'development_level',
+                    'isotach',
+                    'quadrant',
+                    'radius_for_NEQ',
+                    'radius_for_SEQ',
+                    'radius_for_SWQ',
+                    'radius_for_NWQ',
+                    'background_pressure',
+                    'radius_of_last_closed_isobar',
+                    'radius_of_maximum_winds',
+                    'name',
+                    'direction',
+                    'speed',
+                ]
 
-            atcf = self.atcf
-            if isinstance(atcf, io.BytesIO):
-                # test if Gzip file
-                atcf.seek(0)  # rewind
-                if atcf.read(2) == b'\x1f\x8b':
+                atcf = self.atcf
+                if isinstance(atcf, io.BytesIO):
+                    # test if Gzip file
                     atcf.seek(0)  # rewind
-                    atcf = gzip.GzipFile(fileobj=atcf)
-                else:
-                    atcf.seek(0)  # rewind
-
-            start_date = self.start_date
-            # Only accept request record type or
-            # BEST track or OFCL (official) advisory by default
-            allowed_record_types = self.record_type
-            if allowed_record_types is None:
-                allowed_record_types = ['BEST', 'OFCL']
-            records = []
-
-            for line_index, line in enumerate(atcf):
-                line = line.decode('UTF-8').split(',')
-
-                record = {
-                    'basin': line[0],
-                    'storm_number': line[1].strip(' '),
-                }
-
-                record['record_type'] = line[4].strip(' ')
-
-                if record['record_type'] not in allowed_record_types:
-                    continue
-
-                # computing the actual datetime based on record_type
-                if record['record_type'] == 'BEST':
-                    # Add minutes line to base datetime
-                    minutes = line[3].strip(' ')
-                    if minutes == "":
-                        minutes = '00'
-                    record['datetime'] = parse_date(line[2].strip(' ') + minutes)
-                else:
-                    # Add validation time to base datetime
-                    minutes = '00'
-                    record['datetime'] = parse_date(line[2].strip(' ') + minutes)
-                    if start_date is not None:
-                        # Only keep records where base date == start time for advisories
-                        if start_date != record['datetime']:
-                            continue
-                    validation_time = int(line[5].strip(' '))
-                    record['datetime'] = record['datetime'] + timedelta(hours=validation_time)
-
-                latitude = line[6]
-                if 'N' in latitude:
-                    latitude = float(latitude.strip('N '))
-                elif 'S' in latitude:
-                    latitude = float(latitude.strip('S ')) * -1
-                latitude *= 0.1
-                record['latitude'] = latitude
-
-                longitude = line[7]
-                if 'E' in longitude:
-                    longitude = float(longitude.strip('E ')) * 0.1
-                elif 'W' in longitude:
-                    longitude = float(longitude.strip('W ')) * -0.1
-                record['longitude'] = longitude
-
-                record.update(
-                    {
-                        'max_sustained_wind_speed': float(line[8].strip(' ')),
-                        'central_pressure': float(line[9].strip(' ')),
-                        'development_level': line[10].strip(' '),
-                    }
-                )
-
-                try:
-                    record['isotach'] = int(line[11].strip(' '))
-                except ValueError:
-                    raise Exception(
-                        'Error: No radial wind information for this storm; '
-                        'parametric wind model cannot be built.'
-                    )
-
-                record.update(
-                    {
-                        'quadrant': line[12].strip(' '),
-                        'radius_for_NEQ': int(line[13].strip(' ')),
-                        'radius_for_SEQ': int(line[14].strip(' ')),
-                        'radius_for_SWQ': int(line[15].strip(' ')),
-                        'radius_for_NWQ': int(line[16].strip(' ')),
-                    }
-                )
-
-                if len(line) > 18:
-                    record.update(
-                        {
-                            'background_pressure': int(line[17].strip(' ')),
-                            'radius_of_last_closed_isobar': int(line[18].strip(' ')),
-                            'radius_of_maximum_winds': int(line[19].strip(' ')),
-                        }
-                    )
-
-                    if len(line) > 23:
-                        record['name'] = line[27].strip(' ')
+                    if atcf.read(2) == b'\x1f\x8b':
+                        atcf.seek(0)  # rewind
+                        atcf = gzip.GzipFile(fileobj=atcf)
                     else:
-                        record['name'] = ''
-                else:
+                        atcf.seek(0)  # rewind
+
+                start_date = self.start_date
+                # Only accept request record type or
+                # BEST track or OFCL (official) advisory by default
+                allowed_record_types = self.record_type
+                if allowed_record_types is None:
+                    allowed_record_types = ['BEST', 'OFCL']
+                records = []
+
+                for line_index, line in enumerate(atcf):
+                    line = line.decode('UTF-8').split(',')
+
+                    record = {
+                        'basin': line[0],
+                        'storm_number': line[1].strip(' '),
+                    }
+
+                    record['record_type'] = line[4].strip(' ')
+
+                    if record['record_type'] not in allowed_record_types:
+                        continue
+
+                    # computing the actual datetime based on record_type
+                    if record['record_type'] == 'BEST':
+                        # Add minutes line to base datetime
+                        minutes = line[3].strip(' ')
+                        if minutes == "":
+                            minutes = '00'
+                        record['datetime'] = parse_date(line[2].strip(' ') + minutes)
+                    else:
+                        # Add validation time to base datetime
+                        minutes = '00'
+                        record['datetime'] = parse_date(line[2].strip(' ') + minutes)
+                        if start_date is not None:
+                            # Only keep records where base date == start time for advisories
+                            if start_date != record['datetime']:
+                                continue
+                        validation_time = int(line[5].strip(' '))
+                        record['datetime'] = record['datetime'] + timedelta(
+                            hours=validation_time
+                        )
+
+                    latitude = line[6]
+                    if 'N' in latitude:
+                        latitude = float(latitude.strip('N '))
+                    elif 'S' in latitude:
+                        latitude = float(latitude.strip('S ')) * -1
+                    latitude *= 0.1
+                    record['latitude'] = latitude
+
+                    longitude = line[7]
+                    if 'E' in longitude:
+                        longitude = float(longitude.strip('E ')) * 0.1
+                    elif 'W' in longitude:
+                        longitude = float(longitude.strip('W ')) * -0.1
+                    record['longitude'] = longitude
+
                     record.update(
                         {
-                            'background_pressure': records[-1]['background_pressure'],
-                            'radius_of_last_closed_isobar': records[-1][
-                                'radius_of_last_closed_isobar'
-                            ],
-                            'radius_of_maximum_winds': records[-1]['radius_of_maximum_winds'],
-                            'name': records[-1]['name'],
+                            'max_sustained_wind_speed': float(line[8].strip(' ')),
+                            'central_pressure': float(line[9].strip(' ')),
+                            'development_level': line[10].strip(' '),
                         }
                     )
 
-                records.append(record)
+                    try:
+                        record['isotach'] = int(line[11].strip(' '))
+                    except ValueError:
+                        raise Exception(
+                            'Error: No radial wind information for this storm; '
+                            'parametric wind model cannot be built.'
+                        )
 
-            if len(records) == 0:
-                raise ValueError(f'no records found with type(s) "{allowed_record_types}"')
+                    record.update(
+                        {
+                            'quadrant': line[12].strip(' '),
+                            'radius_for_NEQ': int(line[13].strip(' ')),
+                            'radius_for_SEQ': int(line[14].strip(' ')),
+                            'radius_for_SWQ': int(line[15].strip(' ')),
+                            'radius_for_NWQ': int(line[16].strip(' ')),
+                        }
+                    )
 
-            self.__dataframe = DataFrame.from_records(data=records, columns=columns)
+                    if len(line) > 18:
+                        record.update(
+                            {
+                                'background_pressure': int(line[17].strip(' ')),
+                                'radius_of_last_closed_isobar': int(line[18].strip(' ')),
+                                'radius_of_maximum_winds': int(line[19].strip(' ')),
+                            }
+                        )
+
+                        if len(line) > 23:
+                            record['name'] = line[27].strip(' ')
+                        else:
+                            record['name'] = ""
+                    else:
+                        record.update(
+                            {
+                                'background_pressure': records[-1]['background_pressure'],
+                                'radius_of_last_closed_isobar': records[-1][
+                                    'radius_of_last_closed_isobar'
+                                ],
+                                'radius_of_maximum_winds': records[-1][
+                                    'radius_of_maximum_winds'
+                                ],
+                                'name': records[-1]['name'],
+                            }
+                        )
+
+                    records.append(record)
+
+                if len(records) == 0:
+                    raise ValueError(f'no records found with type(s) "{allowed_record_types}"')
+
+                dataframe = DataFrame.from_records(data=records, columns=columns)
+
+            self.__dataframe = dataframe
             self.__previous_configuration = configuration
 
         # if location values have changed, recompute velocity
@@ -661,7 +706,14 @@ class VortexForcing:
         if _found_start_date is False:
             raise Exception(f'No data within mesh bounding box for storm {self.storm_id}.')
 
-    def plot_track(self, axis: Axes = None, show: bool = False, color: str = 'k', **kwargs):
+    def plot_track(
+        self,
+        axis: Axes = None,
+        show: bool = False,
+        color: str = 'k',
+        coastline: bool = True,
+        **kwargs,
+    ):
         kwargs.update({'color': color})
         if axis is None:
             fig = pyplot.figure()
@@ -678,7 +730,103 @@ class VortexForcing:
                 )
         if show:
             axis.axis('scaled')
-        plot_coastline(axis, show)
+        if bool(coastline) is True:
+            plot_coastline(axis, show)
+
+    def isotachs(self, wind_speed: float, segments: int = 91) -> [Polygon]:
+        """
+        Get the isotach at the given speed at every time in the dataset.
+
+        :param wind_speed: wind speed to extract (in knots)
+        :param segments: number of discretization points per quadrant
+        :return: list of isotachs as polygons
+        """
+
+        ## Collect the attributes needed from the forcing to generate swath
+        data = self.data[self.data['isotach'] == wind_speed]
+
+        # convert quadrant radii from nautical miles to meters
+        quadrants = ['radius_for_NEQ', 'radius_for_NWQ', 'radius_for_SWQ', 'radius_for_SEQ']
+        data[quadrants] *= 1852.0  # nautical miles to meters
+
+        geodetic = Geod(ellps='WGS84')
+
+        ## Generate overall swath based on the desired isotach
+        polygons = []
+        for index, row in data.iterrows():
+            # get the starting angle range for NEQ based on storm direction
+            rot_angle = 360 - row['direction']
+            start_angle = 0 + rot_angle
+            end_angle = 90 + rot_angle
+
+            # append quadrants in counter-clockwise direction from NEQ
+            arcs = []
+            for quadrant in quadrants:
+                # enter the angle range for this quadrant
+                theta = numpy.linspace(start_angle, end_angle, segments)
+                # move angle to next quadrant
+                start_angle = start_angle + 90
+                end_angle = end_angle + 90
+                # skip if quadrant radius is zero
+                if row[quadrant] <= 1.0:
+                    continue
+                # make the coordinate list for this quadrant
+                ## entering origin
+                coords = [row[['longitude', 'latitude']].tolist()]
+                # using forward geodetic (origin,angle,dist)
+                for az12 in theta:
+                    lont, latt, backaz = geodetic.fwd(
+                        lons=row['longitude'],
+                        lats=row['latitude'],
+                        az=az12,
+                        dist=row[quadrant],
+                    )
+                    coords.append((lont, latt))
+                ## start point equals last point
+                coords.append(coords[0])
+                # enter quadrant as new polygon
+                arcs.append(Polygon(coords))
+            polygons.append(ops.unary_union(arcs))
+        return polygons
+
+    def wind_swath(
+        self, isotach: int, segments: int = 91, plot_swath: bool = False,
+    ) -> Polygon:
+        """
+        extracts the wind swath of the BestTrackForcing class object as a Polygon object
+
+        :param isotach: the wind swath to extract (34-kt, 50-kt, or 64-kt)
+        :param segments: number of discretization points per quadrant (default = 91)
+        :param plot_swath: plot the swath before returning? (default = False)
+        """
+
+        # parameter
+        # isotach should be one of 34, 50, 64
+        valid_isotach_values = [34, 50, 64]
+        assert (
+            isotach in valid_isotach_values
+        ), f'`isotach` value in `get_wind_swath` must be one of {valid_isotach_values}'
+
+        isotachs = self.isotachs(wind_speed=isotach, segments=segments)
+
+        convex_hulls = []
+        for index in range(len(isotachs) - 1):
+            convex_hulls.append(
+                ops.unary_union([isotachs[index], isotachs[index + 1]]).convex_hull
+            )
+
+        # get the union of polygons
+        swath = ops.unary_union(convex_hulls)
+
+        if plot_swath:
+            plot_polygons(isotachs)
+            plot_polygon(swath)
+            pyplot.suptitle(
+                f'{self.storm_id} - isotach {isotach} kt ({self.start_date} - {self.end_date})'
+            )
+            pyplot.show()
+
+        return swath
 
     def __generate_record_numbers(self):
         record_number = [1]
@@ -697,19 +845,20 @@ class VortexForcing:
                 return date
 
     def __copy__(self) -> 'VortexForcing':
-        instance = self.__class__(
+        return self.__class__(
             storm=self.dataframe.copy(),
             start_date=self.start_date,
             end_date=self.end_date,
             file_deck=self.file_deck,
             record_type=self.record_type,
         )
-        instance.storm_id = self.storm_id
-        return instance
+
+    def __eq__(self, other: 'VortexForcing') -> bool:
+        return numpy.all(self.dataframe == other.dataframe)
 
     @staticmethod
     def __compute_velocity(data: DataFrame) -> DataFrame:
-        """ Output has units of meters per second. """
+        """Output has units of meters per second."""
 
         geodetic = Geod(ellps='WGS84')
 
@@ -757,6 +906,9 @@ class VortexForcing:
     def from_fort22(
         cls, fort22: PathLike, start_date: datetime = None, end_date: datetime = None,
     ) -> 'VortexForcing':
+        filename = None
+        if pathlib.Path(fort22).exists():
+            filename = fort22
         return cls(
             storm=read_atcf(fort22),
             start_date=start_date,
@@ -764,12 +916,16 @@ class VortexForcing:
             file_deck=None,
             mode=None,
             record_type=None,
+            filename=filename,
         )
 
     @classmethod
     def from_atcf_file(
         cls, atcf: PathLike, start_date: datetime = None, end_date: datetime = None,
     ) -> 'VortexForcing':
+        filename = None
+        if pathlib.Path(atcf).exists():
+            filename = atcf
         return cls(
             storm=atcf,
             start_date=start_date,
@@ -777,6 +933,7 @@ class VortexForcing:
             file_deck=None,
             mode=None,
             record_type=None,
+            filename=filename,
         )
 
 
@@ -789,6 +946,7 @@ class BestTrackForcing(VortexForcing, WindForcing):
         start_date: datetime = None,
         end_date: datetime = None,
         mode: Mode = None,
+        filename: PathLike = None,
     ) -> 'BestTrackForcing':
         if nws is None:
             nws = 20
@@ -809,6 +967,7 @@ class BestTrackForcing(VortexForcing, WindForcing):
             file_deck=FileDeck.b,
             mode=mode,
             record_type='BEST',
+            filename=filename,
         )
         WindForcing.__init__(self, nws=nws, interval_seconds=interval_seconds)
 
@@ -835,11 +994,12 @@ class BestTrackForcing(VortexForcing, WindForcing):
         summary = '\n'.join(f)
         if output is not None:
             if not isinstance(output, pathlib.Path):
-                path = pathlib.Path(output)
-            if path.exists() and overwrite is False:
-                raise Exception('File exist, set overwrite=True to allow overwrite.')
-            with open(path, 'w+') as fh:
-                fh.write(summary)
+                output = pathlib.Path(output)
+            if overwrite or not output.exists():
+                with open(output, 'w+') as fh:
+                    fh.write(summary)
+            else:
+                logger.warning(f'skipping existing file "{output}"')
         return summary
 
     @property
@@ -854,15 +1014,6 @@ class BestTrackForcing(VortexForcing, WindForcing):
         assert NWS in [8, 19, 20]
         self.__NWS = int(NWS)
 
-    @property
-    def WTIMINC(self) -> str:
-        return (
-            f'{self.start_date:%Y %m %d %H} '
-            f'{self.data["storm_number"].iloc[0]} '
-            f'{self.BLADj} '
-            f'{self.geofactor}'
-        )
-
     @property
     def BLADj(self) -> float:
         try:
@@ -898,6 +1049,9 @@ class BestTrackForcing(VortexForcing, WindForcing):
         start_date: datetime = None,
         end_date: datetime = None,
     ) -> 'BestTrackForcing':
+        filename = None
+        if pathlib.Path(fort22).exists():
+            filename = fort22
         return cls(
             storm=read_atcf(fort22),
             start_date=start_date,
@@ -905,6 +1059,7 @@ class BestTrackForcing(VortexForcing, WindForcing):
             nws=nws,
             interval_seconds=interval_seconds,
             mode=None,
+            filename=filename,
         )
 
     @classmethod
@@ -916,6 +1071,9 @@ class BestTrackForcing(VortexForcing, WindForcing):
         start_date: datetime = None,
         end_date: datetime = None,
     ) -> 'BestTrackForcing':
+        filename = None
+        if pathlib.Path(atcf).exists():
+            filename = atcf
         return cls(
             storm=atcf,
             start_date=start_date,
@@ -923,6 +1081,7 @@ class BestTrackForcing(VortexForcing, WindForcing):
             nws=nws,
             interval_seconds=interval_seconds,
             mode=None,
+            filename=filename,
         )
 
 
@@ -994,63 +1153,62 @@ def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
 def get_atcf_entry(
     year: int, basin: str = None, storm_number: int = None, storm_name: str = None,
 ) -> Series:
-    entry = None
+    url = 'ftp://ftp.nhc.noaa.gov/atcf/archive/storm.table'
 
     try:
-        url = 'ftp://ftp.nhc.noaa.gov/atcf/archive/storm.table'
         storm_table = read_csv(url, header=None)
-    except:
-        storm_table = None
+    except URLError:
+        raise ConnectionError(f'cannot connect to "{url}"')
 
-    if storm_table is not None:
-        if basin is not None and storm_number is not None:
-            row = storm_table[
-                (storm_table[1] == f'{basin.upper():>3}')
-                & (storm_table[7] == storm_number)
-                & (storm_table[8] == int(year))
-            ]
-        elif storm_name is not None:
-            row = storm_table[
-                (storm_table[0] == f'{storm_name.upper():>10}') & (storm_table[8] == int(year))
-            ]
-        else:
-            raise ValueError('need either storm name or basin + storm number')
+    if basin is not None and storm_number is not None:
+        row = storm_table[
+            (storm_table[1] == f'{basin.upper():>3}')
+            & (storm_table[7] == storm_number)
+            & (storm_table[8] == int(year))
+        ]
+    elif storm_name is not None:
+        row = storm_table[
+            (storm_table[0] == f'{storm_name.upper():>10}') & (storm_table[8] == int(year))
+        ]
+    else:
+        raise ValueError('need either storm name or basin + storm number')
 
-        if len(row) > 0:
-            entry = list(row.iterrows())[0][1]
-
-    return entry
+    if len(row) > 0:
+        return list(row.iterrows())[0][1]
 
 
 def get_atcf_id(storm_name: str, year: int) -> str:
-    entry = get_atcf_entry(storm_name=storm_name, year=year)
-    if entry is None:
-        return None
-    else:
-        return entry[20].strip()
+    return get_atcf_entry(storm_name=storm_name, year=year)[20].strip()
 
 
 def get_atcf_file(storm_id: str, file_deck: FileDeck = None, mode: Mode = None) -> io.BytesIO:
-    url = atcf_url(file_deck=file_deck, storm_id=storm_id, mode=mode).replace('ftp://', '')
+    url = atcf_url(file_deck=file_deck, storm_id=storm_id, mode=mode).replace('ftp://', "")
     logger.info(f'Downloading storm data from {url}')
 
     hostname, filename = url.split('/', 1)
 
     handle = io.BytesIO()
 
-    ftp = ftplib.FTP(hostname, 'anonymous', '')
-    ftp.encoding = 'utf-8'
-    ftp.retrbinary(f'RETR {filename}', handle.write)
+    try:
+        ftp = ftplib.FTP(hostname, 'anonymous', "")
+        ftp.encoding = 'utf-8'
+        ftp.retrbinary(f'RETR {filename}', handle.write)
+    except socket.gaierror:
+        raise ConnectionError(f'cannot connect to {hostname}')
 
     return handle
 
 
 def read_atcf(track: PathLike) -> DataFrame:
     try:
-        with open(track) as track_file:
-            track = track_file.readlines()
-    except:
+        if not isinstance(track, TextIO):
+            track = open(track)
+        track = track.readlines()
+    except FileNotFoundError:
+        # check if the entire track file was passed as a string
         track = str(track).splitlines()
+        if len(track) == 1:
+            raise
 
     data = {
         'basin': [],
@@ -1118,8 +1276,8 @@ def read_atcf(track: PathLike) -> DataFrame:
         row_data['radius_of_maximum_winds'] = convert_value(
             row[19], to_type=int, round_digits=0,
         )
-        row_data['direction'] = row[25]
-        row_data['speed'] = row[26]
+        row_data['direction'] = convert_value(row[25], to_type=int)
+        row_data['speed'] = convert_value(row[26], to_type=int)
         row_data['name'] = row[27]
 
         for key, value in row_data.items():
@@ -1182,9 +1340,9 @@ def atcf_storm_ids(file_deck: FileDeck = None, mode: Mode = None) -> [str]:
     elif not isinstance(file_deck, FileDeck):
         file_deck = convert_value(file_deck, FileDeck)
 
-    url = atcf_url(file_deck=file_deck, mode=mode).replace('ftp://', '')
+    url = atcf_url(file_deck=file_deck, mode=mode).replace('ftp://', "")
     hostname, directory = url.split('/', 1)
-    ftp = ftplib.FTP(hostname, 'anonymous', '')
+    ftp = ftplib.FTP(hostname, 'anonymous', "")
 
     filenames = [
         filename for filename, metadata in ftp.mlsd(directory) if metadata['type'] == 'file'
