@@ -1,3 +1,4 @@
+from collections import Mapping
 import logging
 import pathlib
 
@@ -18,7 +19,7 @@ class NodalAttributes:
     def __str__(self):
         fort13 = [
             f'{self._fort14.description} nodal attributes',
-            f'{len(self._fort14.nodes.id)}',
+            f'{len(self._fort14.nodes.index)}',
             f'{len(self.get_attribute_names())}',
         ]
 
@@ -40,9 +41,7 @@ class NodalAttributes:
             for i, values in enumerate(
                 attribute['values'][attribute['non_default_indexes'], :]
             ):
-                node_id = self._fort14.nodes.get_id_by_index(
-                    attribute['non_default_indexes'][i]
-                )
+                node_id = self._fort14.nodes.index[attribute['non_default_indexes'][i]]
                 line = [f'{node_id}']
                 for value in values:
                     line.append(f'{value}')
@@ -164,7 +163,7 @@ class NodalAttributes:
 
     def import_fort13(self, fort13):
         fort13 = parse_fort13(fort13)
-        if fort13.pop('NumOfNodes') != len(self._fort14.nodes.id):
+        if fort13.pop('NumOfNodes') != len(self._fort14.nodes.index):
             raise Exception('fort.13 file does not match the mesh.')
         self._AGRID = fort13.pop('AGRID')
         for attribute, data in fort13.items():
@@ -175,7 +174,8 @@ class NodalAttributes:
                 (self._fort14.values.size, np.asarray(data['defaults']).flatten().size),
                 np.nan,
             )
-            for i, idx in enumerate(data['indexes']):
+            for i, node_id in enumerate(data['node_id']):
+                idx = self.fort14.nodes.index.get_loc(int(node_id))
                 for j, value in enumerate(values[i, :].tolist()):
                     full_values[idx, j] = value
             idxs = np.where(np.isnan(full_values).all(axis=1))[0]
@@ -185,21 +185,31 @@ class NodalAttributes:
             # converts from column major to row major, leave it column major.
             # if full_values.shape[1] == 1:
             #     full_values = full_values.flatten()
-            self.add_attribute(attribute, data['units'])
+            if attribute not in self._attributes:
+                self.add_attribute(attribute, data['units'])
+            else:
+                logger.warning(f'overwriting existing attribute "{attribute}"')
             self.set_attribute(attribute, full_values)
 
     def write(self, path, overwrite=False):
         if path is not None:
-            path = pathlib.Path(path)
-            if path.is_file() and not overwrite:
-                msg = 'File exists, pass overwrite=True to allow overwrite.'
-                raise Exception(msg)
-            else:
+            if not isinstance(path, pathlib.Path):
+                path = pathlib.Path(path)
+            if overwrite or not path.exists():
                 with open(path, 'w') as f:
                     f.write(str(self))
+            else:
+                logger.warning(f'skipping existing file "{path}"')
         else:
             print(str(self))
 
+    @property
+    def fort14(self):
+        return self._fort14
+
+    def __eq__(self, other: 'NodalAttributes') -> bool:
+        return equal_mappings(self._attributes, other._attributes)
+
 
 def parse_fort13(path):
     fort13 = {}
@@ -218,22 +228,36 @@ def parse_fort13(path):
             fort13[attribute_name] = {
                 'units': units,
                 'defaults': defaults,
-                'indexes': [],
+                'node_id': [],
             }
             i += 1
         for i in range(NAttr):
             attribute_name = f.readline().strip()
             numOfNodes = int(f.readline())
-            values = np.zeros((NP, len(fort13[attribute_name]['defaults'])))
-            values[:] = np.nan
-            j = 0
-            while j < numOfNodes:
-                str = f.readline().split()
-                index = int(str[0]) - 1
-                fort13[attribute_name]['indexes'].append(index)
-                node_values = [float(x) for x in str[1:]]
-                values[index, :] = node_values
-                j += 1
+            # values = np.zeros((NP, len(fort13[attribute_name]["defaults"])))
+            values = np.full((NP, len(fort13[attribute_name]['defaults'])), np.nan)
+            for j in range(numOfNodes):
+                line = f.readline().split()
+                fort13[attribute_name]['node_id'].append(line[0])
+                values[j, :] = [float(x) for x in line[1:]]
             values[np.where(np.isnan(values[:, 0])), :] = fort13[attribute_name]['defaults']
             fort13[attribute_name]['values'] = values
         return fort13
+
+
+def equal_mappings(mapping_1: Mapping, mapping_2: Mapping) -> bool:
+    for key, value_1 in mapping_1.items():
+        if key in mapping_2:
+            value_2 = mapping_2[key]
+            if isinstance(value_1, Mapping):
+                if isinstance(value_2, Mapping):
+                    if not equal_mappings(value_1, value_2):
+                        return False
+                else:
+                    return False
+            elif np.any(value_1 != value_2):
+                return False
+        else:
+            return False
+    else:
+        return True
