@@ -23,24 +23,20 @@ class BaseBoundaries:
     @property
     def ids(self):
         if not hasattr(self, '_ids'):
-            self._ids = list(self._data.keys())
+            self._ids = range(len(self._data))
         return self._ids
 
     @property
     def indexes(self):
         if not hasattr(self, '_indexes'):
-            self._indexes = list()
-            for data in self._data.values():
-                self._indexes.append(
-                    list(map(self._mesh.nodes.get_index_by_id, data['node_id']))
-                )
+            self._indexes = [self._mesh.nodes.index[data['node_id']] for data in self._data]
         return np.array(self._indexes)
 
     @property
     def node_id(self):
         if not hasattr(self, '_node_id'):
             self._node_id = list()
-            for data in self._data.values():
+            for data in self._data:
                 self._node_id.append(data['node_id'])
         return self._node_id
 
@@ -48,10 +44,12 @@ class BaseBoundaries:
     def gdf(self):
         if not hasattr(self, '_gdf'):
             data = []
-            for i, (id, boundary) in enumerate(self._data.items()):
+            for i, boundary in enumerate(self._data):
                 data.append(
                     {
-                        'geometry': LineString(self._mesh.coords[self.indexes[i], :]),
+                        'geometry': LineString(
+                            self._mesh.coords.loc[self.indexes[i], :].values
+                        ),
                         'key': f'{boundary.get("ibtype")}:{id}',
                         'indexes': self.indexes[i],
                         **boundary,
@@ -60,6 +58,9 @@ class BaseBoundaries:
             self._gdf = gpd.GeoDataFrame(data, crs=self._mesh.crs)
         return self._gdf
 
+    def __eq__(self, other: 'BaseBoundaries') -> bool:
+        return self._data == other._data
+
 
 class OceanBoundaries(BaseBoundaries):
     pass
@@ -78,21 +79,6 @@ class InflowBoundaries(BaseBoundaries):
 
 
 class BarrierBaseBoundaries(BaseBoundaries):
-    @property
-    def indexes(self):
-        if not hasattr(self, '_indexes'):
-            self._indexes = list()
-            for data in self._data.values():
-                self._indexes.append(
-                    np.array(
-                        [
-                            list(map(self._mesh.nodes.get_index_by_id, geom))
-                            for geom in data['node_id']
-                        ]
-                    )
-                )
-        return self._indexes
-
     @property
     def gdf(self):
         if not hasattr(self, '_gdf'):
@@ -103,8 +89,8 @@ class BarrierBaseBoundaries(BaseBoundaries):
                     {
                         'geometry': MultiLineString(
                             [
-                                LineString(self._mesh.coords[front_face, :]),
-                                LineString(self._mesh.coords[back_face, :]),
+                                LineString(self._mesh.coords.iloc[front_face, :].values),
+                                LineString(self._mesh.coords.iloc[back_face, :].values),
                             ]
                         ),
                         'key': f'{boundary.get("ibtype")}:{id}',
@@ -207,6 +193,9 @@ class Fort14Boundaries:
                     boundaries.update({len(boundaries) + 1: {'ibtype': ibtype, **bdata,}})
         return boundaries
 
+    def __eq__(self, other: 'Fort14Boundaries') -> bool:
+        return self._data == other._data
+
 
 class Fort14(Grd):
     """
@@ -222,20 +211,26 @@ class Fort14(Grd):
     @classmethod
     def open(cls, path, crs=None):
         _grd = grd.read(path, crs=crs)
-        _grd['nodes'] = {id: (coords, -val) for id, (coords, val) in _grd['nodes'].items()}
+        _grd['nodes'].iloc[:, 2:] *= -1
         return cls(**_grd)
 
+    def write(self, path, overwrite=False, format='fort.14'):
+        if format in ['fort.14']:
+            _grd = self.to_dict()
+            nodes = self.nodes
+            nodes.iloc[:, 2:] *= -1
+            _grd['nodes'] = nodes
+
+            grd.write(
+                grd=_grd, path=path, overwrite=overwrite,
+            )
+        else:
+            super().write(path=path, overwrite=overwrite, format=format)
+
     def to_dict(self, boundaries=True):
         _grd = super().to_dict()
         if boundaries is True:
-            _grd.update(
-                {
-                    'nodes': {
-                        id: (coord, -val) for id, (coord, val) in self.nodes.to_dict().items()
-                    },
-                    'boundaries': self.boundaries.to_dict(),
-                }
-            )
+            _grd.update({'nodes': self.nodes, 'boundaries': self.boundaries.to_dict()})
         return _grd
 
     @figure
@@ -252,9 +247,9 @@ class Fort14(Grd):
         **kwargs,
     ):
         if vmin is None:
-            vmin = np.min(self.values)
+            vmin = np.min(self.values.values)
         if vmax is None:
-            vmax = np.max(self.values)
+            vmax = np.max(self.values.values)
         kwargs.update(**get_topobathy_kwargs(self.values, vmin, vmax))
         kwargs.pop('col_val')
         levels = kwargs.pop('levels')
@@ -311,3 +306,6 @@ class Fort14(Grd):
     @property
     def culvert_boundaries(self):
         return self.boundaries.culvert
+
+    def __eq__(self, other: 'Fort14') -> bool:
+        return super().__eq__(other) and self.boundaries == other.boundaries
